use std::ffi::{CStr, c_char};
use std::sync::{OnceLock, RwLock};

mod lua;
mod plugin;
mod rpc;
mod stingray_sdk;

use log::error;
use plugin::Plugin;
use stingray_sdk::{GetApiFunction, LoggingApi, LuaApi, PluginApi, PluginApiID};

/// The name that the plugin is registered to the engine as.
/// Must be unique across all plugins.
pub const PLUGIN_NAME: &CStr = c"dt_p2p";
/// The module that Lua functions are assigned to.
pub const MODULE_NAME: &CStr = c"dt_p2p";

const CHANNEL_BUFFER_SIZE: usize = 100;
const EVENTS_PER_TICK: usize = 25;

static PLUGIN: OnceLock<RwLock<Plugin>> = OnceLock::new();
static LOGGER: OnceLock<LoggingApi> = OnceLock::new();
static LUA: OnceLock<LuaApi> = OnceLock::new();

/// A macro to make accessing global statics a little more convenient.
#[macro_export]
macro_rules! global {
    ($name:ident) => {{
        if cfg!(debug_assertions) {
            $name.get().expect("global has not been initialized")
        } else {
            unsafe { $name.get().unwrap_unchecked() }
        }
    }};
}

/// A macro to make accessing the plugin instance as mutable a little more convenient.
#[macro_export]
macro_rules! plugin_mut {
    () => {{
        let lock = $crate::PLUGIN.get().and_then(|l| l.write().ok());
        if cfg!(debug_assertions) {
            lock.expect("failed to acquire lock on plugin global")
        } else {
            unsafe { lock.unwrap_unchecked() }
        }
    }};
}

#[unsafe(no_mangle)]
pub extern "C" fn get_name() -> *const c_char {
    PLUGIN_NAME.as_ptr()
}

#[unsafe(no_mangle)]
pub extern "C" fn setup_game(get_engine_api: GetApiFunction) {
    let logger = LOGGER.get_or_init(|| LoggingApi::get(get_engine_api));
    log::set_logger(logger).expect("Failed to set global logger.");
    log::set_max_level(log::LevelFilter::Info);

    let _ = LUA.get_or_init(|| LuaApi::get(get_engine_api));

    let plugin = match Plugin::new() {
        Ok(plugin) => plugin,
        Err(err) => {
            error!("Failed to initialize plugin:\n{:?}", err);
            return;
        }
    };

    plugin.setup_game();
    PLUGIN.set(RwLock::new(plugin));
}

#[unsafe(no_mangle)]
pub extern "C" fn shutdown_game() {
    let plugin = plugin_mut!();
    plugin.shutdown_game();
}

#[unsafe(no_mangle)]
pub extern "C" fn update_game(dt: f32) {
    let mut plugin = plugin_mut!();
    plugin.update_game(dt);
}

#[unsafe(no_mangle)]
pub extern "C" fn get_plugin_api(id: PluginApiID) -> *mut PluginApi {
    if id == PluginApiID::PLUGIN_API_ID {
        let api = PluginApi {
            get_name: Some(get_name),
            setup_game: Some(setup_game),
            update_game: Some(update_game),
            shutdown_game: Some(shutdown_game),
            ..Default::default()
        };

        Box::into_raw(Box::new(api))
    } else {
        std::ptr::null_mut()
    }
}
