use std::collections::HashMap;
use std::panic::catch_unwind;

use bstr::BStr;
use bstr::ByteSlice as _;
use color_eyre::Result;
use color_eyre::eyre;
use color_eyre::eyre::Context as _;

use crate::plugin::Identifier;
use crate::plugin::Plugin;
use crate::stingray_sdk::LUA_REGISTRYINDEX;
use crate::stingray_sdk::{LuaState, LuaType, lua_State};
use crate::{LUA, global, plugin_mut};

pub const NAMESPACE_SEPARATOR: char = '.';

static RPC_CALLBACK_KEY: &str = "dt-p2p.rpc-callbacks";

// As a minor optimization, only `catch_unwind` in debug mode.
#[cfg(debug_assertions)]
fn lua_wrapper<F>(l: *mut lua_State, f: F) -> i32
where
    F: FnOnce(&mut Plugin, &LuaState) -> Result<i32> + std::panic::UnwindSafe,
{
    let lua = LuaState::new(l, global!(LUA));

    let res = catch_unwind(|| {
        let mut plugin = plugin_mut!();
        f(&mut plugin, &lua)
    });
    match res {
        Ok(Ok(i)) => i,
        Ok(Err(err)) => lua.lib_error(format!("{:?}", err)),
        Err(err) => lua.lib_error(format!("{:?}", err)),
    }
}

#[cfg(not(debug_assertions))]
fn lua_wrapper<F>(l: *mut lua_State, f: F) -> i32
where
    F: FnOnce(&mut Plugin, &LuaState) -> Result<i32>,
{
    let lua = LuaState::new(l, global!(LUA));
    let plugin = plugin_mut!();

    let res = f(&mut plugin, &lua);
    match res {
        Ok(i) => i,
        Err(err) => lua.lib_error(format!("{:?}", err)),
    }
}

/// Returns the namespace and name for a RPC.
///
/// The namespace may either be part of the name, separated by `NAMESPACE_SEPARATOR`
/// or omitted, in which case the mod name is used as fallback.
fn get_identifier<'a>(lua: &'a LuaState) -> Result<Identifier<'a>> {
    lua.getfield(1, c"name");
    if !lua.isstring(-1) {
        eyre::bail!("bad argument #1, not a mod object");
    }

    let mod_name = lua.tostring(-1);
    let name = lua.tostring(2);

    let (namespace, name) = name
        .split_once_str(&[NAMESPACE_SEPARATOR as u8])
        .map(|(namespace, name)| (BStr::new(namespace), BStr::new(name)))
        .unwrap_or((mod_name, name));

    Ok(Identifier::new(namespace, name))
}

// Called with `(mod: table, name: string, opts: table)`
pub extern "C" fn create_rpc(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        lua.getfield(1, c"version");
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isstring(2) {
            eyre::bail!("bad argument #2, expected a string");
        }

        if !lua.istable(3) {
            eyre::bail!("bad argument #3, expected a table");
        }

        let _version = match lua.r#type(-1) {
            LuaType::Nil => {
                eyre::bail!("bad argument #1, mod object doesn't provide a 'version' field");
            }
            LuaType::Number => lua.tostring(-1),
            LuaType::String => {
                let version = lua.tostring(-1);
                // TODO: parse as date or semver-like
                version
            }
            x => {
                eyre::bail!("invalid type {} for 'version' field", x)
            }
        };

        let id = get_identifier(lua).wrap_err("failed to determine RPC name")?;
        plugin.create_rpc(id);

        Ok(0)
    })
}

type LuaMap<'a> = HashMap<LuaKey<'a>, LuaValue<'a>>;

#[derive(bincode::BorrowDecode, bincode::Encode, PartialEq)]
enum LuaKey<'a> {
    String(&'a [u8]),
    Number(f64),
}

impl<'a> Eq for LuaKey<'a> {}

impl<'a> std::hash::Hash for LuaKey<'a> {
    fn hash<H: std::hash::Hasher>(&self, state: &mut H) {
        match self {
            LuaKey::String(s) => s.hash(state),
            LuaKey::Number(n) => (*n as u128).hash(state),
        }
    }
}

#[derive(bincode::BorrowDecode, bincode::Encode)]
pub enum LuaValue<'a> {
    Nil,
    Boolean(bool),
    Number(f64),
    String(&'a [u8]),
    Table(LuaMap<'a>),
}

impl<'a> LuaValue<'a> {
    fn parse_table(lua: &'a LuaState, idx: i32) -> Result<Self> {
        let mut values = LuaMap::default();
        //
        // `nil` as initial key tells `lua_next` to pick the first key.
        lua.pushnil();

        loop {
            if lua.next(idx) <= 0 {
                break;
            }

            let key = match lua.r#type(-2) {
                LuaType::Number => LuaKey::Number(lua.tonumber(-2)),
                LuaType::String => LuaKey::String(lua.tostring(-2)),
                t => eyre::bail!("Unsupported type {} for args table key", t),
            };

            let value = Self::parse(lua, -1)?;

            values.insert(key, value);

            // Pop the value, but keep the key so `lua_next` knows where to continue
            lua.pop();
        }

        // Pop the last key
        lua.pop();

        Ok(LuaValue::Table(values))
    }

    pub fn parse(lua: &'a LuaState, idx: i32) -> Result<Self> {
        match lua.r#type(idx) {
            LuaType::None | LuaType::Nil => Ok(LuaValue::Nil),
            LuaType::Boolean => Ok(LuaValue::Boolean(lua.toboolean(-1))),
            LuaType::Number => Ok(LuaValue::Number(lua.tonumber(-1))),
            LuaType::String => Ok(LuaValue::String(lua.tostring(-1))),
            LuaType::Table => Self::parse_table(lua, -1),
            x => eyre::bail!("Unknown or unsupported Lua type {}", x),
        }
    }

    pub fn push(&self, lua: &LuaState) {
        match self {
            LuaValue::Nil => lua.pushnil(),
            LuaValue::Boolean(bool) => lua.pushboolean(*bool),
            LuaValue::Number(num) => lua.pushnumber(*num),
            LuaValue::String(s) => lua.pushstring(*s),
            LuaValue::Table(table) => Self::push_table(lua, table),
        }
    }

    fn push_table(lua: &LuaState, table: &LuaMap) {
        let (narr, nrec) = table.iter().fold((0, 0), |(narr, nrec), (k, _)| match k {
            LuaKey::String(_) => (narr, nrec + 1),
            LuaKey::Number(_) => (narr + 1, nrec),
        });

        lua.createtable(narr, nrec);
        let tbl_index = lua.gettop();

        for (k, v) in table.iter() {
            match k {
                LuaKey::String(s) => lua.pushstring(*s),
                LuaKey::Number(num) => lua.pushnumber(*num),
            }

            v.push(lua);

            lua.settable(tbl_index);
        }
    }
}

// Called with `(mod: table, name: string, args: table)`
pub extern "C" fn send_rpc(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isstring(2) {
            eyre::bail!("bad argument #2, expected a string");
        }

        if !lua.istable(3) {
            eyre::bail!("bad argument #3, expected a table");
        }

        let id = get_identifier(lua).wrap_err("failed to determine RPC name")?;
        let args = LuaValue::parse_table(lua, 3).wrap_err("Failed to parse args table")?;

        plugin.send_rpc(id, args).wrap_err("Failed to send RPC")?;

        Ok(0)
    })
}

// Called with `(mod: table, name: string, callback: Fn(args))`
pub extern "C" fn receive_rpc(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isstring(2) {
            eyre::bail!("bad argument #2, expected a string");
        }

        if !lua.isfunction(3) {
            eyre::bail!("bad argument #3, expected a function");
        }

        let id = get_identifier(lua).wrap_err("failed to determine RPC name")?;

        // We will utilize Lua's _references_ to store the callback functions. Since we're doing
        // the bookkeeping on the Rust side, we don't need to worry about any structured layout on
        // the Lua side.

        lua.pushvalue(3);
        let fn_ref = lua.lib_ref(LUA_REGISTRYINDEX);

        plugin.add_rpc_listener(id, fn_ref);

        Ok(0)
    })
}

// Called with `(mod: table, callback: Fn(peer_id))`
pub extern "C" fn on_peer_connected(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isfunction(2) {
            eyre::bail!("bad argument #2, expected a function");
        }

        lua.getfield(1, c"name");
        if !lua.isstring(-1) {
            eyre::bail!("bad argument #1, not a mod object");
        }

        let mod_name = lua.tostring(-1);
        lua.pushvalue(3);
        let fn_ref = lua.lib_ref(LUA_REGISTRYINDEX);

        plugin.add_connect_listener(mod_name, fn_ref);

        Ok(0)
    })
}

// Called with `(mod: table, callback: Fn(peer_id))`
pub extern "C" fn on_peer_disconnected(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.istable(1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        if !lua.isfunction(2) {
            eyre::bail!("bad argument #2, expected a function");
        }

        lua.getfield(1, c"name");
        if !lua.isstring(-1) {
            eyre::bail!("bad argument #1, expected a mod object");
        }

        let mod_name = lua.tostring(-1);
        lua.pushvalue(3);
        let fn_ref = lua.lib_ref(LUA_REGISTRYINDEX);

        plugin.add_disconnect_listener(mod_name, fn_ref);

        Ok(0)
    })
}

// Called with `(dt: float)`
pub extern "C" fn update(l: *mut lua_State) -> i32 {
    lua_wrapper(l, |plugin, lua| {
        if !lua.isnumber(1) {
            eyre::bail!("bad argument #1, expected a number");
        }

        let dt = lua.tonumber(1);
        plugin.lua_update(dt, lua)?;

        Ok(0)
    })
}
