use std::collections::HashMap;
use std::hash::{DefaultHasher, Hash as _, Hasher as _};
use std::thread;
use std::thread::JoinHandle;

use bstr::{BStr, BString, ByteSlice};
use color_eyre::eyre::Context as _;
use color_eyre::{Result, eyre};
use libp2p::futures::StreamExt;
use libp2p::gossipsub::{AllowAllSubscriptionFilter, Event, IdentityTransform, TopicHash};
use libp2p::swarm::SwarmEvent;
use libp2p::{gossipsub, noise, tcp, yamux};
use tokio::sync::mpsc::Receiver;
use tokio::sync::mpsc::error::TryRecvError;
use tokio::sync::mpsc::{self, UnboundedSender};

use crate::lua::{LuaValue, NAMESPACE_SEPARATOR};
use crate::rpc::{RPC, RPCMessage};
use crate::stingray_sdk::{LUA_REGISTRYINDEX, LuaRef, LuaState};
use crate::{CHANNEL_BUFFER_SIZE, EVENTS_PER_TICK, lua};
use crate::{LUA, MODULE_NAME, global};

#[derive(Clone, Copy, Debug)]
pub struct Identifier<'a> {
    name: &'a BStr,
    namespace: &'a BStr,
}

impl<'a> Identifier<'a> {
    pub fn new(namespace: &'a BStr, name: &'a BStr) -> Self {
        Self { namespace, name }
    }

    pub(crate) fn to_topic(self) -> TopicHash {
        TopicHash::from_raw(format!("{}", self))
    }
}

impl std::fmt::Display for Identifier<'_> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}{}{}", self.namespace, NAMESPACE_SEPARATOR, self.name)
    }
}

#[derive(Debug)]
struct NamespacedMap<T> {
    map: HashMap<BString, HashMap<BString, T>>,
}

impl<T> NamespacedMap<T> {
    pub fn insert(&mut self, id: Identifier, val: T) {
        if !self.map.contains_key(id.namespace) {
            self.map.insert(id.namespace.to_owned(), Default::default());
        }

        let map = self
            .map
            .get_mut(id.namespace)
            .expect("entry is verified to exist");
        map.insert(id.name.to_owned(), val);
    }

    pub fn get(&self, id: &Identifier) -> Option<&T> {
        self.map.get(id.namespace).and_then(|map| map.get(id.name))
    }
}

impl<T> Default for NamespacedMap<T> {
    fn default() -> Self {
        Self {
            map: Default::default(),
        }
    }
}

enum Task {}

pub(crate) struct Plugin {
    swarm_thread: JoinHandle<Result<()>>,
    event_rx: Receiver<SwarmEvent<Event>>,
    send_tx: UnboundedSender<RPCMessage>,
    rpcs: HashMap<TopicHash, RPC>,
    connect_listeners: HashMap<TopicHash, LuaRef>,
    disconnect_listeners: HashMap<TopicHash, LuaRef>,
}

impl Plugin {
    pub fn new() -> Result<Self> {
        let mut swarm = libp2p::SwarmBuilder::with_new_identity()
            .with_tokio()
            .with_tcp(
                tcp::Config::default(),
                noise::Config::new,
                yamux::Config::default,
            )
            .wrap_err("Failed to configure transport")?
            .with_behaviour(|key| {
                let message_id_fn = |msg: &gossipsub::Message| {
                    let mut s = DefaultHasher::new();
                    msg.data.hash(&mut s);
                    gossipsub::MessageId::from(s.finish().to_string())
                };

                let config = gossipsub::ConfigBuilder::default()
                    .validation_mode(gossipsub::ValidationMode::Strict)
                    .message_id_fn(message_id_fn)
                    .build()
                    .wrap_err("Failed to create gossipsub config")?;

                // TODO: Benchmark if compression would be beneficial.
                let behavior: gossipsub::Behaviour<IdentityTransform, AllowAllSubscriptionFilter> =
                    gossipsub::Behaviour::new(
                        gossipsub::MessageAuthenticity::Signed(key.clone()),
                        config,
                    )?;

                Ok(behavior)
            })
            .wrap_err("Failed to configure behavior")?
            .build();

        swarm
            .listen_on(
                "/ipv4/0.0.0.0/tcp/0"
                    .parse()
                    .expect("listen address invalid"),
            )
            .wrap_err("failed to listen on IPv4 socket")?;

        // TODO: Connect to first hard coded relay server(s), which should then
        // fan out to online peers.
        // Maybe a cheap server (e.g. AWS free tier instance) is already enough to handle all
        // session negotiations. If not, then a system is needed where each mod user also becomes
        // available for that, and the hosted server merely acts as load balancer between them.
        //
        // swarm.dial(RELAY_SERVER)

        let (event_tx, event_rx) = mpsc::channel::<SwarmEvent<Event>>(CHANNEL_BUFFER_SIZE);
        // Since the `Sender` will be on the sync side, and we want to avoid blocking, we make this
        // unbounded to limit the chance for it to actually block.
        let (send_tx, rpc_rx) = mpsc::unbounded_channel::<RPCMessage>();

        let swarm_thread = thread::Builder::new()
            .name("p2p-swarm".into())
            .spawn(move || {
                let rt = tokio::runtime::Builder::new_current_thread()
                    .build()
                    .wrap_err("Failed to create tokio runtime")?;

                let mut swarm = swarm;
                let mut rpc_rx = rpc_rx;

                rt.block_on(async move {
                    loop {
                        tokio::select! {
                            event = swarm.select_next_some() => {
                                event_tx.send(event).await.wrap_err("Failed to queue event")?;
                            }
                            msg = rpc_rx.recv() => {
                                let Some(msg) = msg else {
                                    eyre::bail!("RPC queue closed prematurely");
                                };
                                swarm.behaviour_mut().publish(msg.topic, msg.data).wrap_err("Failed to send RPC")?;
                            }
                        }
                    }
                })
            })
            .wrap_err("Failed to create p2p-swarm thread")?;

        Ok(Self {
            swarm_thread,
            event_rx,
            send_tx,
            rpcs: Default::default(),
            connect_listeners: Default::default(),
            disconnect_listeners: Default::default(),
        })
    }

    pub fn setup_game(&self) {
        let lua = global!(LUA);
        lua.add_module_function(MODULE_NAME, c"update", lua::update);

        lua.add_module_function(MODULE_NAME, c"create_rpc", lua::create_rpc);
        lua.add_module_function(MODULE_NAME, c"send_rpc", lua::send_rpc);
        lua.add_module_function(MODULE_NAME, c"receive_rpc", lua::receive_rpc);

        // TODO: Handle mod reloads
        lua.add_module_function(MODULE_NAME, c"on_peer_connected", lua::on_peer_connected);
        lua.add_module_function(
            MODULE_NAME,
            c"on_peer_disconnected",
            lua::on_peer_disconnected,
        );

        log::info!("Plugin initialized");
    }

    pub fn shutdown_game(&self) {
        // TODO: Find a way to send a close command, and make it blocking.
    }

    // WARN: Due to how the instance of `Plugin` is currently stored and
    // accessed by Lua functions, things will likely break if we attempt to get a `lua_State` here
    // and trigger out own Lua functions.
    pub fn update_game(&mut self, _dt: f32) {}

    pub fn lua_update(&mut self, _dt: f64, lua: &LuaState) -> Result<()> {
        if self.swarm_thread.is_finished() {
            eyre::bail!("p2p-swarm thread terminated prematurely",);
            // TODO: Move `self.swarm_thread` into a data structure that I can move it out of here,
            // so that I can call `.join` and get the error it ended with.
        }

        // Limit the amount of events we handle per tick to avoid potential large spikes in
        // frame time.
        // TODO: Add a system to monitor the amount of incoming vs processed events, and add
        // warnings when this cannot catch up.
        // TODO: Check if it might become necessary to create a more elaborate system that can
        // change this limit dynamically based on the amount of incoming messages.
        // This must be able to eventually catch up.
        for _ in 0..EVENTS_PER_TICK {
            let event = match self.event_rx.try_recv() {
                Ok(event) => event,
                Err(TryRecvError::Empty) => break,
                Err(TryRecvError::Disconnected) => {
                    eyre::bail!("p2p-swarm channel disconnected prematurely");
                }
            };

            match event {
                SwarmEvent::Behaviour(Event::Message {
                    propagation_source,
                    message_id,
                    message: msg,
                }) => {
                    log::debug!("Received message {message_id} from {propagation_source}");
                    let Some(rpc) = self.rpcs.get(&msg.topic) else {
                        log::error!("Receive message for unknown topic {}", msg.topic);
                        continue;
                    };

                    let Ok((args, _)) = bincode::borrow_decode_from_slice::<LuaValue<'_>, _>(
                        &msg.data,
                        bincode::config::standard(),
                    ) else {
                        log::error!(
                            "Failed to decode data for message {message_id} on topic {}",
                            msg.topic
                        );
                        continue;
                    };

                    args.push(lua);
                    let args_index = lua.gettop();

                    for fn_ref in rpc.listeners() {
                        lua.rawgeti(LUA_REGISTRYINDEX, *fn_ref);
                        lua.pushvalue(args_index);
                        lua.pcall(1, 0).wrap_err_with(|| {
                            format!("Failed to call listener for RPC '{}'", msg.topic)
                        })?;
                    }
                }
                SwarmEvent::Behaviour(Event::GossipsubNotSupported { peer_id }) => {
                    log::warn!(
                        "Peer {peer_id} does not support Gossipsub. Are they really a dt-p2p user?"
                    );
                }
                SwarmEvent::ConnectionEstablished {
                    peer_id,
                    connection_id,
                    endpoint: _,
                    num_established,
                    concurrent_dial_errors: _,
                    established_in,
                } => {
                    log::info!(
                        "Connection {} established with {} in {} ms. Total: {}",
                        connection_id,
                        peer_id,
                        established_in.as_millis(),
                        num_established
                    );
                    // TODO: Establish mods shared with this peer
                    // TODO: Call `on_peer_connected` listeners for the corresponding mods
                }
                SwarmEvent::ConnectionClosed {
                    peer_id,
                    connection_id,
                    endpoint: _,
                    num_established,
                    cause,
                } => {
                    log::info!(
                        "Connection {} with {} closed for reason {:?}. Total: {}",
                        connection_id,
                        peer_id,
                        cause,
                        num_established
                    );
                    // TODO: Call `on_peer_disconnected` listeners for active mods
                    // TODO: Re-establish active mods
                    // With a peer dropping out, the amount of active mods may change if there is
                    // now only a single peer left for a mod.
                }
                SwarmEvent::IncomingConnectionError {
                    connection_id,
                    local_addr: _,
                    send_back_addr: _,
                    error,
                } => {
                    log::error!("Error on incoming connection {connection_id}: {error}");
                }
                SwarmEvent::OutgoingConnectionError {
                    connection_id,
                    peer_id: _,
                    error,
                } => {
                    log::error!("Error on outgoing connection {connection_id}: {error}");
                }
                SwarmEvent::ListenerError { listener_id, error } => {
                    log::error!("Listener {listener_id} failed: {error}");
                }
                SwarmEvent::NewListenAddr {
                    listener_id,
                    address,
                } => {
                    log::debug!("Listening on {address} with ID {listener_id}");
                }
                // TODO: Maybe add tracing information for the events we don't handle
                _ => {}
            }
        }

        Ok(())
    }

    pub(crate) fn create_rpc(&mut self, id: Identifier) {
        let rpc = RPC::new();
        self.rpcs.insert(id.to_topic(), rpc);
    }

    pub(crate) fn send_rpc(&self, id: Identifier, args: LuaValue<'_>) -> Result<()> {
        if !self.rpcs.contains_key(&id.to_topic()) {
            eyre::bail!("RPC '{}' does not exist", id);
        }

        let args = bincode::encode_to_vec(args, bincode::config::standard())
            .wrap_err("Failed to encode RPC args")?;

        let msg = RPCMessage::new(id, args);
        self.send_tx
            .send(msg)
            .wrap_err("Failed to queue RPC call")?;

        Ok(())
    }

    pub(crate) fn add_rpc_listener(&mut self, id: Identifier, fn_ref: i32) -> Result<()> {
        let Some(rpc) = self.rpcs.get_mut(&id.to_topic()) else {
            eyre::bail!("RPC '{}' does not exist", id);
        };

        rpc.add_listener(fn_ref);

        Ok(())
    }

    pub(crate) fn add_connect_listener(&mut self, mod_name: &BStr, fn_ref: i32) {
        let key = TopicHash::from_raw(mod_name.to_str_lossy());
        self.connect_listeners.insert(key, fn_ref);
    }

    pub(crate) fn add_disconnect_listener(&mut self, mod_name: &BStr, fn_ref: i32) {
        let key = TopicHash::from_raw(mod_name.to_str_lossy());
        self.disconnect_listeners.insert(key, fn_ref);
    }
}

impl std::fmt::Debug for Plugin {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_str("PluginApi")
    }
}
