#![allow(non_upper_case_globals)]
#![allow(non_camel_case_types)]
#![allow(non_snake_case)]
#![allow(clippy::type_complexity)]
#![allow(unused)]

mod bindings {
    include!(concat!(env!("OUT_DIR"), "/bindings.rs"));
}

use std::ffi::CStr;
use std::ffi::CString;
use std::os::raw::c_char;
use std::os::raw::c_int;
use std::os::raw::c_void;

pub use bindings::GetApiFunction;
use bindings::LUA_OK;
pub use bindings::PluginApi;
pub use bindings::PluginApiID;
use bindings::lua_CFunction;
use bindings::lua_Integer;
use bindings::lua_Number;
pub use bindings::lua_State;
pub use bindings::{LUA_ENVIRONINDEX, LUA_GLOBALSINDEX, LUA_REGISTRYINDEX};

use bstr::BStr;
use color_eyre::Result;
use color_eyre::eyre;
use log::Level;

impl std::default::Default for PluginApi {
    fn default() -> Self {
        Self {
            version: 69,
            flags: 3,
            setup_game: None,
            update_game: None,
            shutdown_game: None,
            get_name: None,
            loaded: None,
            start_reload: None,
            unloaded: None,
            finish_reload: None,
            setup_resources: None,
            shutdown_resources: None,
            unregister_world: None,
            register_world: None,
            get_hash: None,
            unkfunc13: None,
            unkfunc14: None,
            unkfunc15: None,
            debug_draw: None,
        }
    }
}

#[cfg(debug_assertions)]
fn get_engine_api(f: GetApiFunction, id: PluginApiID) -> *mut c_void {
    let f = f.expect("'GetApiFunction' is always passed by the engine");
    unsafe { f(id as u32) }
}

#[cfg(not(debug_assertions))]
fn get_engine_api(f: GetApiFunction, id: PluginApiID) -> *mut c_void {
    // `Option::unwrap` still generates several instructions in
    // optimized code.
    let f = unsafe { f.unwrap_unchecked() };

    unsafe { f(id as u32) }
}

pub struct LoggingApi {
    info: unsafe extern "C" fn(*const c_char, *const c_char),
    warning: unsafe extern "C" fn(*const c_char, *const c_char),
    error: unsafe extern "C" fn(*const c_char, *const c_char),
}

impl LoggingApi {
    pub fn get(f: GetApiFunction) -> Self {
        let api = unsafe {
            let api = get_engine_api(f, PluginApiID::LOGGING_API_ID);
            api as *mut bindings::LoggingApi
        };

        unsafe {
            Self {
                info: (*api).info.unwrap_unchecked(),
                warning: (*api).warning.unwrap_unchecked(),
                error: (*api).error.unwrap_unchecked(),
            }
        }
    }

    pub fn info(&self, system: impl Into<Vec<u8>>, message: impl Into<Vec<u8>>) {
        let system = CString::new(system).expect("Invalid CString");
        let message = CString::new(message).expect("Invalid CString");
        unsafe {
            (self.info)(system.as_ptr(), message.as_ptr());
        }
    }

    pub fn warning(&self, system: impl Into<Vec<u8>>, message: impl Into<Vec<u8>>) {
        let system = CString::new(system).expect("Invalid CString");
        let message = CString::new(message).expect("Invalid CString");
        unsafe {
            (self.warning)(system.as_ptr(), message.as_ptr());
        }
    }

    pub fn error(&self, system: impl Into<Vec<u8>>, message: impl Into<Vec<u8>>) {
        let system = CString::new(system).expect("Invalid CString");
        let message = CString::new(message).expect("Invalid CString");
        unsafe {
            (self.error)(system.as_ptr(), message.as_ptr());
        }
    }
}

impl log::Log for LoggingApi {
    fn enabled(&self, metadata: &log::Metadata) -> bool {
        true
    }

    fn log(&self, record: &log::Record) {
        if !self.enabled(record.metadata()) {
            return;
        }

        let line = if let Some(s) = record.args().as_str() {
            format!("[{}] {}", record.level(), s)
        } else {
            format!("[{}] {}", record.level(), record.args())
        };

        self.info(record.target(), line);
    }

    // The engine does not expose a method to flush the log
    fn flush(&self) {}
}

#[repr(i32)]
#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum LuaStatus {
    Ok = bindings::LUA_OK as i32,
    Yield = bindings::LUA_YIELD as i32,
    ErrRun = bindings::LUA_ERRRUN as i32,
    ErrSyntax = bindings::LUA_ERRSYNTAX as i32,
    ErrMem = bindings::LUA_ERRMEM as i32,
    ErrErr = bindings::LUA_ERRERR as i32,
}

#[repr(i32)]
#[derive(Clone, Copy, Debug, PartialEq, Eq)]
pub enum LuaType {
    None = -1,
    Nil = 0,
    Boolean = 1,
    LightUserdata = 2,
    Number = 3,
    String = 4,
    Table = 5,
    Function = 6,
    Userdata = 7,
    Thread = 8,
    Unknown(i32),
}

impl From<i32> for LuaType {
    fn from(value: i32) -> Self {
        match value {
            -1 => Self::None,
            0 => Self::Nil,
            1 => Self::Boolean,
            2 => Self::LightUserdata,
            3 => Self::Number,
            4 => Self::String,
            5 => Self::Table,
            6 => Self::Function,
            7 => Self::Userdata,
            8 => Self::Thread,
            _ => Self::Unknown(value),
        }
    }
}

impl std::fmt::Display for LuaType {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Self::None => write!(f, "None"),
            Self::Nil => write!(f, "Nil"),
            Self::Boolean => write!(f, "Boolean"),
            Self::LightUserdata => write!(f, "LightUserdata"),
            Self::Number => write!(f, "Number"),
            Self::String => write!(f, "String"),
            Self::Table => write!(f, "Table"),
            Self::Function => write!(f, "Function"),
            Self::Userdata => write!(f, "Userdata"),
            Self::Thread => write!(f, "Thread"),
            Self::Unknown(val) => write!(f, "Unkown({})", val),
        }
    }
}

pub type LuaRef = i32;

pub struct LuaApi {
    add_module_function: unsafe extern "C" fn(*const c_char, *const c_char, lua_CFunction),
    createtable: unsafe extern "C" fn(*mut lua_State, i32, i32),
    getfield: unsafe extern "C" fn(*mut lua_State, i32, *const c_char),
    gettable: unsafe extern "C" fn(*mut lua_State, i32),
    gettop: unsafe extern "C" fn(*mut lua_State) -> i32,
    isnumber: unsafe extern "C" fn(*mut lua_State, i32) -> c_int,
    isnil: unsafe extern "C" fn(*mut lua_State, i32) -> c_int,
    isstring: unsafe extern "C" fn(*mut lua_State, i32) -> c_int,
    istable: unsafe extern "C" fn(*mut lua_State, i32) -> c_int,
    lib_argerror: unsafe extern "C" fn(*mut lua_State, i32, *const c_char) -> i32,
    lib_checklstring: unsafe extern "C" fn(*mut lua_State, i32, *mut usize) -> *const c_char,
    lib_error: unsafe extern "C" fn(*mut lua_State, *const c_char, ...) -> i32,
    lib_ref: unsafe extern "C" fn(*mut lua_State, i32) -> i32,
    next: unsafe extern "C" fn(*mut lua_State, i32) -> i32,
    objlen: unsafe extern "C" fn(*mut lua_State, i32) -> usize,
    pcall: unsafe extern "C" fn(*mut lua_State, i32, i32, i32) -> i32,
    pop: unsafe extern "C" fn(*mut lua_State),
    pushboolean: unsafe extern "C" fn(*mut lua_State, i32),
    pushinteger: unsafe extern "C" fn(*mut lua_State, lua_Integer),
    pushnil: unsafe extern "C" fn(*mut lua_State),
    pushnumber: unsafe extern "C" fn(*mut lua_State, lua_Number),
    pushstring: unsafe extern "C" fn(*mut lua_State, *const c_char),
    pushvalue: unsafe extern "C" fn(*mut lua_State, i32),
    rawgeti: unsafe extern "C" fn(*mut lua_State, i32, i32),
    r#type: unsafe extern "C" fn(*mut lua_State, i32) -> i32,
    settable: unsafe extern "C" fn(*mut lua_State, i32),
    toboolean: unsafe extern "C" fn(*mut lua_State, i32) -> i32,
    tolstring: unsafe extern "C" fn(*mut lua_State, i32, *mut usize) -> *const c_char,
    tonumber: unsafe extern "C" fn(*mut lua_State, i32) -> f64,
}

impl LuaApi {
    pub fn get(f: GetApiFunction) -> Self {
        let api = unsafe {
            let api = get_engine_api(f, PluginApiID::LUA_API_ID);
            api as *mut bindings::LuaApi
        };

        unsafe {
            Self {
                add_module_function: (*api).add_module_function.unwrap_unchecked(),
                createtable: (*api).createtable.unwrap_unchecked(),
                getfield: (*api).getfield.unwrap_unchecked(),
                gettable: (*api).gettable.unwrap_unchecked(),
                gettop: (*api).gettop.unwrap_unchecked(),
                isnumber: (*api).isnumber.unwrap_unchecked(),
                isnil: (*api).isnil.unwrap_unchecked(),
                isstring: (*api).isstring.unwrap_unchecked(),
                istable: (*api).istable.unwrap_unchecked(),
                lib_argerror: (*api).lib_argerror.unwrap_unchecked(),
                lib_checklstring: (*api).lib_checklstring.unwrap_unchecked(),
                lib_error: (*api).lib_error.unwrap_unchecked(),
                lib_ref: (*api).lib_ref.unwrap_unchecked(),
                next: (*api).next.unwrap_unchecked(),
                objlen: (*api).objlen.unwrap_unchecked(),
                pcall: (*api).pcall.unwrap_unchecked(),
                pop: (*api).pop.unwrap_unchecked(),
                pushboolean: (*api).pushboolean.unwrap_unchecked(),
                pushinteger: (*api).pushinteger.unwrap_unchecked(),
                pushnil: (*api).pushnil.unwrap_unchecked(),
                pushnumber: (*api).pushnumber.unwrap_unchecked(),
                pushstring: (*api).pushstring.unwrap_unchecked(),
                pushvalue: (*api).pushvalue.unwrap_unchecked(),
                rawgeti: (*api).rawgeti.unwrap_unchecked(),
                r#type: (*api).type_.unwrap_unchecked(),
                settable: (*api).settable.unwrap_unchecked(),
                toboolean: (*api).toboolean.unwrap_unchecked(),
                tolstring: (*api).tolstring.unwrap_unchecked(),
                tonumber: (*api).tonumber.unwrap_unchecked(),
            }
        }
    }

    pub fn add_module_function(
        &self,
        module: impl AsRef<CStr>,
        name: impl AsRef<CStr>,
        cb: extern "C" fn(*mut lua_State) -> i32,
    ) {
        unsafe {
            (self.add_module_function)(module.as_ref().as_ptr(), name.as_ref().as_ptr(), Some(cb))
        }
    }
}

/// A wrapper that combines a lua_State and a LuaApi to make calling functions more convenient.
pub struct LuaState<'a> {
    l: *mut lua_State,
    api: &'a LuaApi,
}

impl<'a> LuaState<'a> {
    pub fn new(l: *mut lua_State, api: &'a LuaApi) -> Self {
        Self { l, api }
    }

    pub fn gettop(&self) -> i32 {
        unsafe { (self.api.gettop)(self.l) }
    }

    pub fn pushvalue(&self, idx: i32) {
        unsafe { (self.api.pushvalue)(self.l, idx) }
    }

    pub fn isnil(&self, idx: i32) -> bool {
        let is_nil = unsafe { (self.api.isnil)(self.l, idx) };
        is_nil > 0
    }

    pub fn isnumber(&self, idx: i32) -> bool {
        let is_number = unsafe { (self.api.isnumber)(self.l, idx) };
        is_number > 0
    }

    pub fn isstring(&self, idx: i32) -> bool {
        let is_string = unsafe { (self.api.isstring)(self.l, idx) };
        is_string > 0
    }

    pub fn isfunction(&self, idx: i32) -> bool {
        let is_fn = unsafe { (self.api.isfunction)(self.l, idx) };
        is_fn > 0
    }

    pub fn istable(&self, idx: i32) -> bool {
        let is_table = unsafe { (self.api.istable)(self.l, idx) };
        is_table > 0
    }

    pub fn r#type(&self, idx: i32) -> LuaType {
        let t = unsafe { (self.api.r#type)(self.l, idx) };
        t.into()
    }

    pub fn tonumber(&self, idx: i32) -> f64 {
        unsafe { (self.api.tonumber)(self.l, idx) }
    }

    pub fn toboolean(&self, idx: i32) -> bool {
        let n = unsafe { (self.api.toboolean)(self.l, idx) };
        n != 0
    }

    pub fn tostring(&self, idx: i32) -> &BStr {
        let bytes = unsafe {
            let mut len = 0usize;
            // `c_char` is `i8` by default, but printable characters are all > 0, so we don't care.
            let c = (self.api.tolstring)(self.l, idx, &mut len) as *const u8;
            std::slice::from_raw_parts(c, len)
        };
        BStr::new(bytes)
    }

    pub fn objlen(&self, idx: i32) -> usize {
        unsafe { (self.api.objlen)(self.l, idx) }
    }

    pub fn pushnil(&self) {
        unsafe { (self.api.pushnil)(self.l) };
    }

    pub fn pushnumber(&self, n: f64) {
        unsafe { (self.api.pushnumber)(self.l, n) };
    }

    pub fn pushinteger(&self, n: isize) {
        unsafe { (self.api.pushinteger)(self.l, n) };
    }

    pub fn pushstring(&self, s: impl Into<Vec<u8>>) {
        let s = CString::new(s).expect("Invalid CString");
        unsafe { (self.api.pushstring)(self.l, s.as_ptr()) };
    }

    pub fn pushboolean(&self, b: bool) {
        unsafe { (self.api.pushboolean)(self.l, b as i32) }
    }

    pub fn gettable(&self, idx: i32) {
        unsafe { (self.api.gettable)(self.l, idx) }
    }

    pub fn getfield(&self, idx: i32, k: impl AsRef<CStr>) {
        unsafe { (self.api.getfield)(self.l, idx, k.as_ref().as_ptr()) }
    }

    pub fn createtable(&self, narr: i32, nrec: i32) {
        unsafe { (self.api.createtable)(self.l, narr, nrec) }
    }

    pub fn settable(&self, idx: i32) {
        unsafe { (self.api.settable)(self.l, idx) }
    }

    pub fn rawgeti(&self, idx: i32, n: i32) {
        unsafe { (self.api.rawgeti)(self.l, idx, n) }
    }

    pub fn call(&self, nargs: i32, nresults: i32) {
        self.pcall(nargs, nresults).unwrap()
    }

    pub fn pcall(&self, nargs: i32, nresults: i32) -> Result<()> {
        // TODO: Re-create the engine's error handler function to populate the stack trace and
        // local variables.
        let res = unsafe { (self.api.pcall)(self.l, nargs, nresults, 0) };
        if res as u32 == LUA_OK {
            return Ok(());
        }

        let err = self.tostring(-1);
        eyre::bail!("pcall failed: {}", err)
    }

    pub fn next(&self, idx: i32) -> i32 {
        unsafe { (self.api.next)(self.l, idx) }
    }

    pub fn lib_argerror(&self, narg: i32, msg: impl AsRef<CStr>) -> i32 {
        unsafe { (self.api.lib_argerror)(self.l, narg, msg.as_ref().as_ptr()) }
    }

    pub fn lib_checklstring(&self, idx: i32) -> &BStr {
        let bytes = unsafe {
            let mut len = 0usize;
            // `c_char` is `i8` by default, but printable characters are all > 0, so we don't care.
            let c = (self.api.lib_checklstring)(self.l, idx, &mut len) as *const u8;
            std::slice::from_raw_parts(c, len)
        };
        BStr::new(bytes)
    }

    // Lua's `luaL_error` does have printf-like formatting capabilities,
    // but since we can just use `format!()`, we don't need to bother handling the varargs here.
    pub fn lib_error(&self, msg: impl Into<Vec<u8>>) -> i32 {
        let s = CString::new(msg).expect("Invalid CString");
        unsafe { (self.api.lib_error)(self.l, s.as_ptr()) }
    }

    pub fn lib_ref(&self, t: i32) -> LuaRef {
        unsafe { (self.api.lib_ref)(self.l, t) }
    }

    // NOTE: This is not like the standard `lua_pop`, it only pops one value at a time.
    // If more values need to be popped, use `settop(-x)`.
    pub fn pop(&self) {
        unsafe { (self.api.pop)(self.l) };
    }
}
